from gobbli.dataset.imdb import IMDBDataset
from gobbli.dataset.newsgroups import NewsgroupsDataset
from gobbli.dataset.trivial import TrivialDataset

__all__ = ["TrivialDataset", "NewsgroupsDataset", "IMDBDataset"]
